/**
    @file tmc_app_usb.c

    @brief Source file for USB functions for TMC Class demo on FT51.
 *

    @version 1.0.0

**/
/*
 * ============================================================================
 * History
 * =======
 *
 * (C) Copyright 2014, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */

/* INCLUDES ************************************************************************/
#include "tmc_app_usb.h"

#if LCD
#include "ft51_lcd.h"
#endif // LCD

/**
 @name ms_timer
 @brief Timer from tmc_app.c Used for attach-detach delay.
 */
#if DFU_DETACH_MODE
extern volatile int16_t ms_timer;
#endif // DFU_DETACH_MODE

/* CONSTANTS ***********************************************************************/

/**
 @name device_descriptor_demo
 @brief Device descriptor for Run Time mode.
 */
__code USB_device_descriptor device_descriptor_demo =
{
    .bLength = 0x12,
    .bDescriptorType = USB_DESCRIPTOR_TYPE_DEVICE,
    .bcdUSB = USB_BCD_VERSION_2_0,          // V2.0
    .bDeviceClass = USB_CLASS_DEVICE,       // Defined in interface
    .bDeviceSubClass = USB_SUBCLASS_DEVICE, // Defined in interface
    .bDeviceProtocol = USB_PROTOCOL_DEVICE, // Defined in interface
    .bMaxPacketSize0 = USB_CONTROL_EP_MAX_PACKET_SIZE,   // 16
    .idVendor = USB_VID_FTDI,   // idVendor: 0x0403 (FTDI)
    .idProduct = USB_PID_DEMO, // idProduct: 0x0feb
    .bcdDevice = 0x0101,        // 1.1
    .iManufacturer = 0x01,      // Manufacturer
    .iProduct = 0x02,           // Product
    .iSerialNumber = 0x03,      // Serial Number
    .bNumConfigurations = 0x01
};

/**
 @name config_descriptor_demo
 @brief Config descriptor for Run Time mode.
 */
//@{
__code struct config_descriptor_demo
{
    USB_configuration_descriptor configuration;
    USB_interface_descriptor interface;
    USB_endpoint_descriptor endpoint_in;
    USB_endpoint_descriptor endpoint_out;
#if DFU
    USB_interface_descriptor dfu_interface;
    USB_dfu_functional_descriptor dfu_functional;
#endif // DFU
};

// Fields marked with * are updated with information in device_config[]
struct config_descriptor_demo config_descriptor_demo =
{
    .configuration.bLength = 0x09,
    .configuration.bDescriptorType = USB_DESCRIPTOR_TYPE_CONFIGURATION,
    .configuration.wTotalLength = sizeof(struct config_descriptor_demo),
#if DFU
    .configuration.bNumInterfaces = 0x02,
#else // DFU
    .configuration.bNumInterfaces = 0x01,
#endif // DFU
    .configuration.bConfigurationValue = 0x01,
    .configuration.iConfiguration = 0x00,
    .configuration.bmAttributes = USB_CONFIG_BMATTRIBUTES_VALUE,
    .configuration.bMaxPower = 0xFA,           // 500mA
    // ---- INTERFACE DESCRIPTOR for Demo ----
    .interface.bLength = 0x09,
    .interface.bDescriptorType = USB_DESCRIPTOR_TYPE_INTERFACE,
    .interface.bInterfaceNumber = 0,
    .interface.bAlternateSetting = 0x00,
    .interface.bNumEndpoints = 0x02,
    .interface.bInterfaceClass = USB_CLASS_APPLICATION, // Application class
    .interface.bInterfaceSubClass = 0x03, // Test and Measurement
    .interface.bInterfaceProtocol = 0x00, // 0x00 = USBTMC
                                          // 0x01 = USBTMC USB488 interface.
    .interface.iInterface = 0x06,         // "FT51A Demo"
    // ---- ENDPOINT DESCRIPTOR for Demo ----
    .endpoint_in.bLength = 0x07,
    .endpoint_in.bDescriptorType = USB_DESCRIPTOR_TYPE_ENDPOINT,
    .endpoint_in.bEndpointAddress = 0x01,
    .endpoint_in.bmAttributes = USB_ENDPOINT_DESCRIPTOR_ATTR_BULK,
    .endpoint_in.wMaxPacketSize = USB_BULK_EP_MAX_PACKET_SIZE,
    .endpoint_in.bInterval = 0x0,
    // ---- ENDPOINT DESCRIPTOR for Demo ----
    .endpoint_out.bLength = 0x07,
    .endpoint_out.bDescriptorType = USB_DESCRIPTOR_TYPE_ENDPOINT,
    .endpoint_out.bEndpointAddress = 0x81,
    .endpoint_out.bmAttributes = USB_ENDPOINT_DESCRIPTOR_ATTR_BULK,
    .endpoint_out.wMaxPacketSize = USB_BULK_EP_MAX_PACKET_SIZE,
    .endpoint_out.bInterval = 0x0,
#if DFU
    // ---- INTERFACE DESCRIPTOR for DFU Interface ----
    .dfu_interface.bLength = 0x09,
    .dfu_interface.bDescriptorType = USB_DESCRIPTOR_TYPE_INTERFACE,
    .dfu_interface.bInterfaceNumber = DFU_USB_INTERFACE_RUNTIME,
    .dfu_interface.bAlternateSetting = 0x00,
    .dfu_interface.bNumEndpoints = 0x00,
    .dfu_interface.bInterfaceClass = USB_CLASS_APPLICATION, // Application Specific Class
    .dfu_interface.bInterfaceSubClass = USB_SUBCLASS_DFU, // Device Firmware Update
    .dfu_interface.bInterfaceProtocol = USB_PROTOCOL_DFU_RUNTIME, // Runtime Protocol
    .dfu_interface.iInterface = 0x05,       // * iInterface: "DFU Interface"
    // ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
    .dfu_functional.bLength = 0x09,
    .dfu_functional.bDescriptorType = USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL,
#if DFU_DETACH_MODE
	.dfu_functional.bmAttributes = USB_DFU_BMATTRIBUTES_CANDNLOAD | USB_DFU_BMATTRIBUTES_WILLDETACH,
#else // DFU_DETACH_MODE
	.dfu_functional.bmAttributes = USB_DFU_BMATTRIBUTES_CANDNLOAD,
#endif // DFU_DETACH_MODE
    .dfu_functional.wDetatchTimeOut = DFU_TIMEOUT, // suggest 8192ms
    .dfu_functional.wTransferSize = DFU_BLOCK_SIZE, // 64 bytes
    .dfu_functional.bcdDfuVersion = USB_BCD_VERSION_DFU_1_1, // DFU Version 1.1
#endif // DFU
};
//@}

#if DFU
/**
 @name device_descriptor_dfumode
 @brief Device descriptor for DFU Mode.
 */
__code USB_device_descriptor device_descriptor_dfumode =
{
    .bLength = 0x12,
    .bDescriptorType = USB_DESCRIPTOR_TYPE_DEVICE,
    .bcdUSB = USB_BCD_VERSION_2_0,          // V2.00
    .bDeviceClass = USB_CLASS_DEVICE,       // Defined in interface
    .bDeviceSubClass = USB_SUBCLASS_DEVICE, // Defined in interface
    .bDeviceProtocol = USB_PROTOCOL_DEVICE, // Defined in interface
    .bMaxPacketSize0 = USB_CONTROL_EP_MAX_PACKET_SIZE,    // 16
    .idVendor = USB_VID_FTDI,   // idVendor: 0x0403 (FTDI)
    .idProduct = DFU_USB_PID_DFUMODE, // idProduct: 0x0fee
    .bcdDevice = 0x0101,        // 1.1
    .iManufacturer = 0x01,      // Manufacturer
    .iProduct = 0x04,           // Product
    .iSerialNumber = 0x03,      // Serial Number
    .bNumConfigurations = 0x01
};
#endif // DFU

#if DFU
/**
 @name config_descriptor_dfumode
 @brief Config descriptor for DFU Mode.
 */
//@{
__code struct config_descriptor_dfumode
{
    USB_configuration_descriptor configuration;
    USB_interface_descriptor dfu_interface;
    USB_dfu_functional_descriptor dfu_functional;
};

// Fields marked with * are updated with information in device_config[]
struct config_descriptor_dfumode config_descriptor_dfumode =
{
    .configuration.bLength = 0x09,
    .configuration.bDescriptorType = USB_DESCRIPTOR_TYPE_CONFIGURATION,
    .configuration.wTotalLength = sizeof(struct config_descriptor_dfumode),
    .configuration.bNumInterfaces = 0x01,
    .configuration.bConfigurationValue = 0x01,
    .configuration.iConfiguration = 0x00,
    .configuration.bmAttributes = USB_CONFIG_BMATTRIBUTES_VALUE,
    .configuration.bMaxPower = 0xFA, // 500mA

    // ---- INTERFACE DESCRIPTOR for DFU Interface ----
    .dfu_interface.bLength = 0x09,
    .dfu_interface.bDescriptorType = USB_DESCRIPTOR_TYPE_INTERFACE,
    .dfu_interface.bInterfaceNumber = DFU_USB_INTERFACE_DFUMODE,
    .dfu_interface.bAlternateSetting = 0x00,
    .dfu_interface.bNumEndpoints = 0x00,
    .dfu_interface.bInterfaceClass = USB_CLASS_APPLICATION, // Application Specific Class
    .dfu_interface.bInterfaceSubClass = USB_SUBCLASS_DFU, // Device Firmware Update
    .dfu_interface.bInterfaceProtocol = USB_PROTOCOL_DFU_DFUMODE, // Runtime Protocol
    .dfu_interface.iInterface = 0x05,
    // ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
    .dfu_functional.bLength = 0x09,
    .dfu_functional.bDescriptorType = USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL,
#if DFU_DETACH_MODE
	.dfu_functional.bmAttributes = USB_DFU_BMATTRIBUTES_CANDNLOAD | USB_DFU_BMATTRIBUTES_WILLDETACH,
#else // DFU_DETACH_MODE
	.dfu_functional.bmAttributes = USB_DFU_BMATTRIBUTES_CANDNLOAD,
#endif // DFU_DETACH_MODE
    .dfu_functional.wDetatchTimeOut = DFU_TIMEOUT, // suggest 8192ms
    .dfu_functional.wTransferSize = DFU_BLOCK_SIZE,     // 64
    .dfu_functional.bcdDfuVersion = USB_BCD_VERSION_DFU_1_1, // DFU Version 1.1
};
#endif
//@}

/**
 @name string_descriptor
 @brief Table of USB String descriptors

 This is placed at a fixed location in the __code section allowing
 up-to 256 bytes of string descriptors to be defined. These can be
 modified or replaced by a utility or binary editor without
 requiring a recompilation of the firmware.
 They are placed at offset 0x100 and reserve 0x100 bytes.
 The data is not stored in __xdata section and is therefore
 regarded as const.
 */
__code __at(STRING_DESCRIPTOR_LOCATION)uint8_t string_descriptor[STRING_DESCRIPTOR_ALLOCATION] =
{
    // String 0 is actually a list of language IDs supported by the real strings.
    4, USB_DESCRIPTOR_TYPE_STRING,
    0x09, 0x04, // 0409 = English (US)

    // String 1 (Manufacturer): "FTDI"
    10, USB_DESCRIPTOR_TYPE_STRING,
    'F', 0x00, 'T', 0x00, 'D', 0x00, 'I', 0x00,

    // String 2 (Product): "FT51A TMC demo"
    30, USB_DESCRIPTOR_TYPE_STRING,
    'F', 0x00, 'T', 0x00, '5', 0x00, '1', 0x00, 'A', 0x00, ' ', 0x00, 'T', 0x00,
    'M', 0x00, 'C', 0x00, ' ', 0x00, 'd', 0x00, 'e', 0x00, 'm', 0x00, 'o', 0x00,

    // String 3 (Serial Number): "NI-VISA-42422-a"
    32, USB_DESCRIPTOR_TYPE_STRING,
    'N', 0x00, 'I', 0x00, '-', 0x00, 'V', 0x00, 'I', 0x00, 'S', 0x00, 'A', 0x00,
    '-', 0x00, '4', 0x00, '2', 0x00, '4', 0x00, '2', 0x00, '2', 0x00, '-', 0x00,
    'a', 0x00,

#if DFU
    // String 4 (DFU Product Name): "FT51A DFU Mode"
    30, USB_DESCRIPTOR_TYPE_STRING,
    'F', 0x00, 'T', 0x00, '5', 0x00, '1', 0x00, 'A', 0x00, ' ', 0x00, 'D', 0x00,
    'F', 0x00, 'U', 0x00, ' ', 0x00, 'M', 0x00, 'o', 0x00, 'd', 0x00, 'e', 0x00,

    // String 5 (Interface Name): "DFU Demo"
    28, USB_DESCRIPTOR_TYPE_STRING,
    'D', 0x00, 'F', 0x00, 'U', 0x00, ' ', 0x00, 'I', 0x00, 'n', 0x00, 't', 0x00,
    'e', 0x00, 'r', 0x00, 'f', 0x00, 'a', 0x00, 'c', 0x00, 'e', 0x00,
#else // DFU
    4, USB_DESCRIPTOR_TYPE_STRING,
    '4', 0x00,
    4, USB_DESCRIPTOR_TYPE_STRING,
    '5', 0x00,
#endif // DFU

    // String 6 (Interface Name): "USBTMC Demo"
    24, USB_DESCRIPTOR_TYPE_STRING,
    'U', 0x00, 'S', 0x00, 'B', 0x00, 'T', 0x00, 'M', 0x00, 'C', 0x00, ' ', 0x00,
    'D', 0x00, 'e', 0x00, 'm', 0x00, 'o', 0x00,

    // END OF STRINGS
    0
};

/* GLOBAL VARIABLES ****************************************************************/
char cmdResp[64-12];
uint8_t cmdRespLen;

/**
 @name bulk_out
 @brief BULK OUT endpoint setup status.
 */
FT51_STATUS bulk_out = FT51_FAILED;
FT51_STATUS bulk_in = FT51_FAILED;

/* LOCAL VARIABLES *****************************************************************/

/* MACROS **************************************************************************/

/* LOCAL FUNCTIONS / INLINES *******************************************************/

FT51_STATUS class_req_cb(USB_device_request *req);
FT51_STATUS standard_req_get_descriptor(USB_device_request *req);
FT51_STATUS standard_req_get_status(USB_device_request *req);
FT51_STATUS standard_req_set_clear_feature(USB_device_request *req);
FT51_STATUS standard_req_cb(USB_device_request *req);
#if 0
FT51_STATUS vendor_req_cb(USB_device_request *req);
void suspend_cb(uint8_t status);
void resume_cb(uint8_t status);
#endif
void reset_cb(uint8_t status);
#if DFU_DETACH_MODE
void usb_detach_attach(void);
#endif // DFU_DETACH_MODE

/**
 @brief      USB class request handler
 @details    Handle class requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and the DFU state machine and status updated as
 required.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bRequest is not
 supported.
 **/
FT51_STATUS class_req_cb(USB_device_request *req)
{
    FT51_STATUS status = FT51_FAILED;
    uint8_t interface = LSB(req->wIndex) & 0x0F;

    // For DFU requests the SETUP packet must be addressed to the
    // the DFU interface on the device.

    // For DFU requests ensure the recipient is an interface...
    if ((req->bmRequestType & USB_BMREQUESTTYPE_RECIPIENT_MASK) ==
    USB_BMREQUESTTYPE_RECIPIENT_INTERFACE)
    {

#if DFU
        // ...and that the interface is the correct Runtime interface
        if (dfu_is_runtime())
#endif // DFU
        {

#if DFU
            if ((interface == DFU_USB_INTERFACE_RUNTIME))
            {
                status = FT51_OK;
                // Handle only DFU_DETATCH, DFU_GETSTATE and DFU_GETSTATUS
                // when in Runtime mode. Table 3.2 DFU_DETACH is mandatory
                // in Runtime mode, DFU_GETSTATE and DFU_GETSTATUS are
                // optional.
                switch (req->bRequest)
                {
                case USB_CLASS_REQUEST_DETACH:
                    dfu_class_req_detach(req->wValue);
#if DFU_DETACH_MODE
                    usb_detach_attach();
#endif // DFU_DETACH_MODE
                    break;
                case USB_CLASS_REQUEST_GETSTATUS:
                    dfu_class_req_getstatus();
                    break;
                case USB_CLASS_REQUEST_GETSTATE:
                    dfu_class_req_getstate();
                    break;
                default:
                    status = FT51_FAILED;
                }
            }
#if TMC
            else
#endif // TMC
#endif // DFU
#if TMC
            {
                status = FT51_OK;
                // Handle only TMC Class interface requests when not
                // in DFU mode.
            	switch ( req->bRequest)
            	{
                default:
                    status = FT51_FAILED;
                    break;
                case USBTMC_GET_CAPABILITIES:
                    tmc_class_req_capabilities(0, 0);
                    break;
                case USBTMC_INITIATE_CLEAR:
                    tmc_class_init_clear();
                    break;
                case USBTMC_CHECK_CLEAR_STATUS:
                    tmc_class_check_clear();
                    break;
            	}
            }
#endif // TMC
        }
#if DFU
        // ...or the correct DFU Mode interface
        else
        {
            if (interface == DFU_USB_INTERFACE_DFUMODE)
            {
                status = FT51_OK;
                // Handle remaining DFU class requests when in DFU Mode.
                // Table 3.2 DFU_DETACH is not supported in DFU Mode.
                switch (req->bRequest)
                {
                case USB_CLASS_REQUEST_DNLOAD:
                    /* For simplicity only program the maximum block size
                     * starting at an address which is an exact multiple
                     * of this. The block size is given in the configuration
                     * descriptor for the DFU interface.
                     */
                    dfu_class_req_download(req->wValue * DFU_BLOCK_SIZE,
                            req->wLength);
                    break;
                case USB_CLASS_REQUEST_GETSTATUS:
                    dfu_class_req_getstatus();
#if DFU_DETACH_MODE
                    if (dfu_is_wait_reset())
                    {
                    	usb_detach_attach();
                    }
#endif // DFU_DETACH_MODE
					break;
                case USB_CLASS_REQUEST_GETSTATE:
                    dfu_class_req_getstate();
                    break;
                case USB_CLASS_REQUEST_CLRSTATUS:
                    dfu_class_req_clrstatus();
                    break;
                case USB_CLASS_REQUEST_ABORT:
                    dfu_class_req_abort();
                    break;
                case USB_CLASS_REQUEST_UPLOAD:
                default:
                    status = FT51_FAILED;                    // Unknown or unsupported request.
                    break;
                }
            }
        }
#endif // DFU
    }
    // Check for Endpoint recipient
#if TMC
    else if((req->bmRequestType & USB_BMREQUESTTYPE_RECIPIENT_MASK) ==
    USB_BMREQUESTTYPE_RECIPIENT_ENDPOINT)
    {
		// TMC class endpoint requests
		status = FT51_OK;
		switch (req->bRequest) {
		default:
			status = FT51_FAILED;
			break;
		case USBTMC_INITIATE_ABORT_BULK_OUT:
			tmc_class_init_abort_bulk_out((req->wValue) & 0xFF, USB_EP_1,
					USB_DIR_OUT);
			break;
		case USBTMC_CHECK_ABORT_BULK_OUT_STATUS:
			tmc_class_check_abort_bulk_out();
			break;
		case USBTMC_INITIATE_ABORT_BULK_IN:
			tmc_class_init_abort_bulk_in((req->wValue) & 0xFF, USB_EP_1,
					USB_DIR_IN);
			break;
		case USBTMC_CHECK_ABORT_BULK_IN_STATUS:
			tmc_class_check_abort_bulk_in();
			break;
		}
	}
#endif // TMC

    return status;
}

/**
 @brief      USB standard request handler for GET_DESCRIPTOR
 @details    Handle standard GET_DESCRIPTOR requests from the host
 application.
 The hValue is parsed and the appropriate device,
 configuration or string descriptor is selected.
 For device and configuration descriptors the DFU
 state machine determines which descriptor to use.
 For string descriptors the lValue selects which
 string from the string_descriptors table to use.
 Strings are obtained from the __code section and
 typecast to a be compatible with __xdata storage.
 The string table is automatically traversed to find
 the correct string and the length is taken from the
 string descriptor.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bmRequestType is not
 supported.
 **/
FT51_STATUS standard_req_get_descriptor(USB_device_request *req)
{
    uint8_t *src = NULL;
    uint16_t length = req->wLength;
    uint8_t hValue = req->wValue >> 8;
    uint8_t lValue = req->wValue & 0x00ff;
    uint8_t i;
    uint8_t slen;

    switch (hValue)
    {
    case USB_DESCRIPTOR_TYPE_DEVICE:

#if DFU
        if (dfu_is_runtime())
#endif // DFU
        {
            src = (char *) &device_descriptor_demo;
        }
#if DFU
        else
        {
            src = (char *) &device_descriptor_dfumode;
        }
#endif // DFU
        if (length > sizeof(USB_device_descriptor)) // too many bytes requested
            length = sizeof(USB_device_descriptor); // Entire structure.
        break;

    case USB_DESCRIPTOR_TYPE_CONFIGURATION:

#if DFU
        if (dfu_is_runtime())
#endif // DFU
        {
            src = (char *) &config_descriptor_demo;
            if (length > sizeof(config_descriptor_demo)) // too many bytes requested
                length = sizeof(config_descriptor_demo); // Entire structure.
        }
#if DFU
        else
        {
            src = (char *) &config_descriptor_dfumode;
            if (length > sizeof(config_descriptor_dfumode)) // too many bytes requested
                length = sizeof(config_descriptor_dfumode); // Entire structure.
        }
#endif // DFU
        break;

    case USB_DESCRIPTOR_TYPE_STRING:

        // Find the nth string in the string descriptor table
        i = 0;
        while ((slen = string_descriptor[i]) > 0)
        {
            // Point to start of string descriptor in __code segment
            // Typecast prevents warning for losing const when usb_transfer
            // is called
            src = (uint8_t *) &string_descriptor[i];
            if (lValue == 0)
            {
                break;
            }
            i += slen;
            lValue--;
        }
        if (lValue > 0)
        {
            // String not found
            return FT51_FAILED;
        }
        // Update the length returned only if it is less than the requested
        // size
        if (length > slen)
        {
            length = slen;
        }

        break;

    default:
        return FT51_FAILED;
    }

    USB_transfer(USB_EP_0, USB_DIR_IN, src, length);

    return FT51_OK;
}

/**
 @brief      USB standard request handler for GET_STATUS
 @details    Handle standard GET_STATUS requests from the host
             application.
             The bmRequestType value is parsed and the appropriate
             action or function is performed.
 @param[in]	 req - USB_device_request structure containing the
             SETUP portion of the request from the host.
 @return	 status - FT51_OK if successful or FT51_FAILED
             if there is an error or the bmRequestType is not
             supported.
 **/
FT51_STATUS standard_req_get_status(USB_device_request *req)
{
    USB_STATE state = USB_get_state();
    USB_ENDPOINT_NUMBER ep_number = LSB(req->wIndex) & 0x0F;
    USB_ENDPOINT_DIR ep_dir = (LSB(req->wIndex) >> 7);
    uint8_t buf[2];

    buf[0] = buf[1] = 0;

    // Get Status for endpoints, only halt status is needed
    if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_DEV_TO_HOST |
            USB_BMREQUESTTYPE_RECIPIENT_ENDPOINT))
    {
        if (((state < CONFIGURED) && (ep_number == 0))
                || (state >= CONFIGURED))
        {
            if (USB_get_ep_stalled(ep_number, ep_dir))
            {
                buf[0] = USB_GET_STATUS_ENDPOINT_HALT;
            }
            USB_transfer(USB_EP_0, USB_DIR_IN, buf, 2);

            return FT51_OK;
        }
    }
    else if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_DEV_TO_HOST |
            USB_BMREQUESTTYPE_RECIPIENT_DEVICE))
    {
        // This will return the self-powered status (currently fixed)
#if USB_SELF_POWERED == 1
        buf[0] = USB_GET_STATUS_DEVICE_SELF_POWERED;
#else // USB_SELF_POWERED
        buf[0] = 0;
#endif // USB_SELF_POWERED

#if USB_REMOTE_WAKEUP == 1
        if (USB_get_remote_wakeup())
        {
            buf[0] |= USB_GET_STATUS_DEVICE_REMOTE_WAKEUP;
        }
#endif // USB_REMOTE_WAKEUP

        USB_transfer(USB_EP_0, USB_DIR_IN, buf, 2);

        return FT51_OK;
    }
    if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_DEV_TO_HOST |
            USB_BMREQUESTTYPE_RECIPIENT_INTERFACE))
    {
        // Must return 2 bytes of zeros for an interface
        USB_transfer(USB_EP_0, USB_DIR_IN, buf, 2);

        return FT51_OK;
    }
    return FT51_FAILED;
}

/**
 @brief      USB standard request handler for CLEAR_FEATURE and SET_FEATURE
 @details    Handle standard CLEAR_FEATURE and SET_FEATURE requests
             from the host application.
             The bmRequestType value is parsed and the appropriate
             action or function is performed.
 @param[in]  req - USB_device_request structure containing the
             SETUP portion of the request from the host.
 @return        status - FT51_OK if successful or FT51_FAILED
             if there is an error or the bmRequestType is not
             supported.
 **/
FT51_STATUS standard_req_set_clear_feature(USB_device_request *req)
{
    USB_ENDPOINT_NUMBER ep_number = LSB(req->wIndex) & 0x0F;
    USB_ENDPOINT_DIR ep_dir = (LSB(req->wIndex) >> 7);
    USB_STATE state = USB_get_state();

    if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_HOST_TO_DEV |
            USB_BMREQUESTTYPE_RECIPIENT_ENDPOINT))
    {
        // Only support the endpoint halt feature for endpoint
        if (req->wValue == USB_FEATURE_ENDPOINT_HALT)
        {
            // Only allow this if the device is configured or the endpoint
            // is zero if the device is not configured.
            if (((state < CONFIGURED) && (ep_number == 0))
                    || (state >= CONFIGURED))
            {
                // Perform a stall or a clear
                if (req->bRequest == USB_REQUEST_CODE_CLEAR_FEATURE)
                {
                    // Feature selector CLEAR
                    USB_clear_endpoint(ep_number, ep_dir);
                }
                else
                {
                    // Feature selector SET
                    USB_stall_endpoint(ep_number, ep_dir);
                }

                USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
                return FT51_OK;
            }
        }
    }
    if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_HOST_TO_DEV |
            USB_BMREQUESTTYPE_RECIPIENT_DEVICE))
    {
#if USB_REMOTE_WAKEUP == 1
        // Only support the remote wakeup feature in this device
        if (req->wValue == USB_FEATURE_DEVICE_REMOTE_WAKEUP)
        {
            // Perform a stall or a clear
            if (req->bRequest == USB_REQUEST_CODE_CLEAR_FEATURE)
            {
                // Feature selector CLEAR
                USB_clear_remote_wakeup();
            }
            else
            {
                // Feature selector SET
                USB_set_remote_wakeup();
            }

            USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
            return FT51_OK;
        }
#endif // USB_REMOTE_WAKEUP
    }
    return FT51_FAILED;
}

/**
 @brief      USB standard request handler
 @details    Handle standard requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and provided to other handlers.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bRequest is not
 supported.
 **/
FT51_STATUS standard_req_cb(USB_device_request *req)
{
    FT51_STATUS status = FT51_FAILED;

    switch (req->bRequest)
    {
    case USB_REQUEST_CODE_GET_STATUS:
        status = standard_req_get_status(req);
        break;

    case USB_REQUEST_CODE_CLEAR_FEATURE:
    case USB_REQUEST_CODE_SET_FEATURE:
        status = standard_req_set_clear_feature(req);
        break;

    case USB_REQUEST_CODE_GET_DESCRIPTOR:
        status = standard_req_get_descriptor(req);
        break;

    case USB_REQUEST_CODE_SET_CONFIGURATION:
        status = USB_req_set_configuration(req);
        break;

    case USB_REQUEST_CODE_GET_CONFIGURATION:
        status = USB_req_get_configuration();
        break;

    case USB_REQUEST_CODE_SET_ADDRESS:
        status =  USB_req_set_address(req);
        break;

    default:
    case USB_REQUEST_CODE_GET_INTERFACE:
    case USB_REQUEST_CODE_SET_INTERFACE:
        // Unknown or unsupported request.
        status = FT51_FAILED;
        break;
    }

    return status;
}

/**
 @brief      USB vendor request handler
 @details    Handle vendor requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and provided to other handlers.
 There are no vendor requests requiring handling in
 this firmware.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bRequest is not
 supported.
 **/
#if 0
FT51_STATUS vendor_req_cb(USB_device_request *req)
{
    (void) req;
    return FT51_FAILED;
}
#endif // 0

/**
 @brief      USB suspend callback
 @details    Called when the USB bus enters the suspend state.
 @param[in]	status - unused.
 @return		N/A
 **/
#if 0
void suspend_cb(uint8_t status)
{
    //char suspMsg[21];
    (void) status;
    lcd_position(0, 15);
    lcd_display("SUSPD", 5);
    return;
}
#endif // 0

/**
 @brief      USB resume callback
 @details    Called when the USB bus enters the resume state
 prior to restating after a suspend.
 @param[in]  status - unused.
 @return     N/S
 **/
#if 0
void resume_cb(uint8_t status)
{
    (void) status;
    lcd_position(0, 15);
    lcd_display("RUN  ", 5);
    return;
}
#endif // 0

/**
 @brief      USB reset callback
 @details    Called when the USB bus enters the reset state.
 @param[in]	status - unused.
 @return		N/A
 **/
void reset_cb(uint8_t status)
{
    (void) status;

#if LCD
    lcd_position(0, 15);
#if DFU
    if (!dfu_is_runtime())
    {
    	lcd_display("DFU  ", 5);
    }
    else
#endif // DFU
    {
    	lcd_display("RUN  ", 5);
    }
#endif // LCD

    USB_free_endpoint(USB_EP_1, USB_DIR_OUT);
    USB_free_endpoint(USB_EP_1, USB_DIR_IN);

#if DFU
    if (dfu_reset())
    {
        // Run downloaded application - reload MTP to shadow RAM
        // Execution effectively stops now and returns to the main()
        // function of the new program in MTP.
        device_revert();
    }
#endif // DFU

    /* Make application specific BULK, INTERRUPT or
     * ISOCHRONOUS endpoints.
     */
    bulk_out = USB_create_endpoint(USB_EP_1, USB_EP_BULK,
            USB_DIR_OUT, USB_BULK_EP_SIZE);
    bulk_in = USB_create_endpoint(USB_EP_1, USB_EP_BULK,
            USB_DIR_IN, USB_BULK_EP_SIZE);

    return;
}

/**
 @brief      USB detach attach sequence
 @details    Called when the USB device wishes to detach from the
 	 	 	 bus then reattach.
 @return		N/A
 **/
#if DFU_DETACH_MODE
void usb_detach_attach(void)
{
	ms_timer = 10;
	while (ms_timer > 0);
	// Disable USB device to force an reset
	USB_finalise();
	ms_timer = 10;
	while (ms_timer > 0);
	// Start the USB device.
	usb_setup();
}
#endif // DFU_DETACH_MODE

/* FUNCTIONS ***********************************************************************/

/**
    @name SCPI_cmds
    @brief Array of commands which device supports.
 **/
__code char *SCPI_cmds[] = {
#ifdef FULL_SCPI
    "*CLS\n",
    "*RST\n",
    "*TST\n",
    "*OPC\n",
    "*OPC?\n",
    "*WAI\n",
    "*ESE\n",
    "*ESE?\n",
    "*ESR?\n",
    "*SRE\n",
    "*SRE?\n",
    "*STB?\n",
#endif
    "*IDN?\n",
#if MEAS_HR
    "*MEAS:hrm?\n",
#endif
#if MEAS_FORCE
    "*MEAS:for?\n",
#endif
#if MEAS_TEMP
    "*MEAS:tmp?\n",
#endif
    NULL,
};

/**
 @brief   \par
 @details
 callbacks
 **/

typedef void (*SCPI_CMD_t)(char * pBuffer);

void scpi_cmd_identify_query(char * pBuffer)
{
    (void)pBuffer;

    strcpy (cmdResp, "FTDI USBTMC Device FT42422:\n");
    cmdRespLen = 28; // return string length
}

#ifdef FULL_SCPI
void scpi_cmd_clear_status(char * pBuffer)
{
    strcpy (cmdResp, "Clear:\n");
    cmdRespLen = 8; //return string length
}

void scpi_cmd_reset(char * pBuffer)
{
    strcpy (cmdResp, "Reset:\n");
    cmdRespLen = 8; //return string length
}

void scpi_cmd_self_test_query(char * pBuffer)
{
    strcpy (cmdResp, "Test:\n");
    cmdRespLen = 7; //return string length
}

void scpi_cmd_operation_complete(char * pBuffer)
{
    strcpy (cmdResp, "Complete:\n");
    cmdRespLen = 11; //return string length
}

void scpi_cmd_operation_complete_query(char * pBuffer)
{
    strcpy (cmdResp, "Complete?:\n");
    cmdRespLen = 12; //return string length
}

void scpi_cmd_wait_to_continue(char * pBuffer)
{
    strcpy (cmdResp, "Continue:\n");
    cmdRespLen = 11; //return string length
}

void scpi_cmd_standard_event_status_enable(char * pBuffer)
{
    strcpy (cmdResp, "Status?:\n");
    cmdRespLen = 10; //return string length
}

void scpi_cmd_standard_event_status_enable_query(char * pBuffer)
{
    strcpy (cmdResp, "Enable?:\n");
    cmdRespLen = 9; //return string length
}

void scpi_cmd_standard_event_status_register_query(char * pBuffer)
{
    strcpy (cmdResp, "Register?:\n");
    cmdRespLen = 12; //return string length
}

void scpi_cmd_service_request_enable(char * pBuffer)
{
    strcpy (cmdResp, "Service:\n");
    cmdRespLen = 11; //return string length
}

void scpi_cmd_service_request_enable_query(char * pBuffer)
{
    strcpy (cmdResp, "Service?:\n");
    cmdRespLen = 12; //return string length
}

void scpi_cmd_read_status_byte_query(char * pBuffer)
{
    strcpy (cmdResp, "Read?:\n");
    cmdRespLen = 8; //return string length
}
#endif

#if MEAS_TEMP
void scpi_cmd_meas_temp(char * pBuffer)
{
	extern uint16_t temperature;

    (void)pBuffer;
    sprintf (cmdResp, "%2u.%2u\n", temperature / 100, temperature % 100);
    cmdRespLen = 7; // return string length
}
#endif

#if MEAS_FORCE
void scpi_cmd_meas_force(char * pBuffer)
{
    extern uint16_t force;

    (void)pBuffer;
    sprintf (cmdResp, "%5u\n", force);
    cmdRespLen = 7; // return string length
}
#endif

#if MEAS_HR
void scpi_cmd_meas_hr(char * pBuffer)
{
    extern uint16_t heartRate;

    (void)pBuffer;
    sprintf (cmdResp, "%5u\n", heartRate);
    cmdRespLen = 7; // return string length
}
#endif

__code SCPI_CMD_t scpi_cmd_list[] =
{
#ifdef FULL_SCPI
    clear_status,
    reset,
    self_test_query,
    operation_complete,
    operation_complete_query,
    wait_to_continue,

    standard_event_status_enable,
    standard_event_status_enable_query,
    standard_event_status_register_query,

    service_request_enable,
    service_request_enable_query,
    read_status_byte_query,
#endif
    scpi_cmd_identify_query, // *IDN?
#if MEAS_HR
    scpi_cmd_meas_hr, // *MEAS:hrm?
#endif
#if MEAS_FORCE
    scpi_cmd_meas_force, // *MEAS:for?
#endif
#if MEAS_TEMP
    scpi_cmd_meas_temp, // *MEAS:tmp?
#endif
    NULL,
};

void tmc_command_cb(char *command)
{
    uint8_t c;

    for (c = 0 ;; c++)
    {
        if (SCPI_cmds[c] == NULL)
        {
            break;
        }
        if ( !strcmp(command, SCPI_cmds[c]) )
        {
            if (scpi_cmd_list[c])
            {
                scpi_cmd_list[c](command);
            }
            break;
        }
    }
}

uint8_t tmc_response_cb(char *response)
{
    uint8_t len;

    strcpy(response, cmdResp);
    len = cmdRespLen;

    // Invalidate response buffer.
    cmdRespLen = 0;

    return len;
}

/* FUNCTIONS ***********************************************************************/

/**
 @brief      Initialise emulation of a USB device
 @details    Sets up callback functions and data structures
 required to implement control endpoints on device.
 @param[in]	ep_size - Endpoint size of control endpoints.
 @return     N/A
 **/
void usb_setup(void)
{
    USB_ctx usb_ctx;
#if TMC
    TMC_ctx tmc_ctx;
#endif // TMC

    memset(&usb_ctx, 0, sizeof(usb_ctx));

    usb_ctx.standard_req = standard_req_cb;
    usb_ctx.class_req = class_req_cb;
#if 0
    usb_ctx.vendor_req = vendor_req_cb;
#endif // 0
#if 0
    usb_ctx.suspend = suspend_cb;
    usb_ctx.resume = resume_cb;
#endif // LCD
    usb_ctx.reset = reset_cb;
    //usb_ctx.lpm = NULL;

    usb_ctx.hub_en.mask.hub_enable = USB_HUB_ENABLED;
    usb_ctx.hub_en.mask.hub_LocalPwrSrc = USB_SELF_POWERED;
    usb_ctx.hub_en.mask.hub_RmtWkUpEn = USB_REMOTE_WAKEUP;

    // Initialise the USB device with a control endpoint size
    // of 8 bytes. This must match the size for bMaxPacketSize0
    // defined in the device descriptor.
    usb_ctx.ep_size = USB_CONTROL_EP_SIZE;

    // Perform USB Device initialisation.
    USB_initialise(&usb_ctx);

#if TMC
    tmc_ctx.message_cmd_cb = tmc_command_cb;
    tmc_ctx.message_rsp_cb = tmc_response_cb;

    // Initialise the TMC interface
    tmc_initialise(&tmc_ctx);
#endif // TMC

    return;
}
